@@ -1,5 +1,5 @@
 use genson_core::normalise::{normalise_values, MapEncoding, NormaliseConfig};
-use genson_core::{infer_json_schema_from_strings, SchemaInferenceConfig};
+use genson_core::{infer_json_schema_from_strings, DebugVerbosity, SchemaInferenceConfig};
 use polars::prelude::*;
 use polars_jsonschema_bridge::deserialise::{schema_to_polars_fields, SchemaFormat};
 use pyo3_polars::derive::polars_expr;
@@ -21,6 +21,12 @@ pub struct GensonKwargs {
     #[serde(default)]
     pub debug: bool,
 
+    #[serde(default)]
+    pub profile: bool,
+
+    #[serde(default)]
+    pub verbosity: DebugVerbosity,
+
     #[serde(default = "default_merge_schemas")]
     pub merge_schemas: bool,
 
@@ -31,9 +37,22 @@ pub struct GensonKwargs {
     #[serde(default = "default_map_threshold")]
     pub map_threshold: usize,
 
+    #[serde(default)]
+    pub map_max_required_keys: Option<usize>,
+
+    /// Enable unification of compatible but non-homogeneous record schemas into maps
+    #[serde(default)]
+    pub unify_maps: bool,
+
+    #[serde(default)]
+    pub no_unify: Vec<String>,
+
     #[serde(default)]
     pub force_field_types: std::collections::HashMap<String, String>,
 
+    #[serde(default = "default_wrap_scalars")]
+    pub wrap_scalars: bool,
+
     /// Whether to emit Avro schema instead of JSON Schema
     #[serde(default)]
     pub avro: bool,
@@ -54,6 +73,9 @@ pub struct GensonKwargs {
     ///   an object with key `"field"`.
     #[serde(default)]
     pub wrap_root: Option<String>,
+
+    #[serde(default = "default_no_root_map")]
+    pub no_root_map: bool,
 }
 
 fn default_map_threshold() -> usize {
@@ -72,6 +94,14 @@ fn default_empty_as_null() -> bool {
     true
 }
 
+fn default_wrap_scalars() -> bool {
+    true
+}
+
+fn default_no_root_map() -> bool {
+    true
+}
+
 fn default_map_encoding() -> MapEncoding {
     MapEncoding::KeyValueEntries
 }
@@ -149,9 +179,17 @@ pub fn infer_json_schema(inputs: &[Series], kwargs: GensonKwargs) -> PolarsResul
                 delimiter: if kwargs.ndjson { Some(b'\n') } else { None },
                 schema_uri: kwargs.schema_uri.clone(),
                 map_threshold: kwargs.map_threshold,
+                map_max_required_keys: kwargs.map_max_required_keys,
+                unify_maps: kwargs.unify_maps,
+                no_unify: kwargs.no_unify.iter().cloned().collect(),
                 force_field_types: kwargs.force_field_types.clone(),
+                wrap_scalars: kwargs.wrap_scalars,
                 avro: kwargs.avro,
                 wrap_root: wrap_root_field.clone(),
+                no_root_map: kwargs.no_root_map,
+                debug: kwargs.debug,
+                profile: kwargs.profile,
+                verbosity: kwargs.verbosity,
             };
 
             let schema_result = infer_json_schema_from_strings(&json_strings, config)
@@ -185,9 +223,17 @@ pub fn infer_json_schema(inputs: &[Series], kwargs: GensonKwargs) -> PolarsResul
                     delimiter: if kwargs.ndjson { Some(b'\n') } else { None },
                     schema_uri: kwargs.schema_uri.clone(),
                     map_threshold: kwargs.map_threshold,
+                    map_max_required_keys: kwargs.map_max_required_keys,
+                    unify_maps: kwargs.unify_maps,
+                    no_unify: kwargs.no_unify.iter().cloned().collect(),
                     force_field_types: kwargs.force_field_types.clone(),
+                    wrap_scalars: kwargs.wrap_scalars,
                     avro: kwargs.avro,
                     wrap_root: wrap_root_field.clone(),
+                    no_root_map: kwargs.no_root_map,
+                    debug: kwargs.debug,
+                    profile: kwargs.profile,
+                    verbosity: kwargs.verbosity,
                 };
 
                 let single_result = infer_json_schema_from_strings(from_ref(json_str), config)
@@ -264,9 +310,17 @@ pub fn infer_polars_schema(inputs: &[Series], kwargs: GensonKwargs) -> PolarsRes
             delimiter: if kwargs.ndjson { Some(b'\n') } else { None },
             schema_uri: kwargs.schema_uri.clone(),
             map_threshold: kwargs.map_threshold,
+            map_max_required_keys: kwargs.map_max_required_keys,
+            unify_maps: kwargs.unify_maps,
+            no_unify: kwargs.no_unify.iter().cloned().collect(),
             force_field_types: kwargs.force_field_types.clone(),
+            wrap_scalars: kwargs.wrap_scalars,
             avro: kwargs.avro,
             wrap_root: wrap_root_field,
+            no_root_map: kwargs.no_root_map,
+            debug: kwargs.debug,
+            profile: kwargs.profile,
+            verbosity: kwargs.verbosity,
         };
 
         let schema_result = infer_json_schema_from_strings(&json_strings, config)
@@ -396,9 +450,17 @@ pub fn normalise_json(inputs: &[Series], kwargs: GensonKwargs) -> PolarsResult<S
         delimiter: if kwargs.ndjson { Some(b'\n') } else { None },
         schema_uri: kwargs.schema_uri.clone(),
         map_threshold: kwargs.map_threshold,
+        map_max_required_keys: kwargs.map_max_required_keys,
+        unify_maps: kwargs.unify_maps,
+        no_unify: kwargs.no_unify.iter().cloned().collect(),
         force_field_types: kwargs.force_field_types.clone(),
+        wrap_scalars: kwargs.wrap_scalars,
         avro: true, // normalisation implies Avro
         wrap_root: wrap_root_field.clone(),
+        no_root_map: kwargs.no_root_map,
+        debug: kwargs.debug,
+        profile: kwargs.profile,
+        verbosity: kwargs.verbosity,
     };
 
     let schema_result = infer_json_schema_from_strings(&json_strings, config)
